import abc

from kupala.application import Kupala
from kupala.container import Container


class Provider(abc.ABC):
    def register(self, app: Kupala) -> None:
        pass

    def bootstrap(self, container: Container) -> None:
        pass



import jinja2
import typing as t
from jinja2.ext import Extension

from kupala.application import Kupala
from kupala.container import Resolver
from kupala.contracts import TemplateRenderer
from kupala.providers import Provider


class JinjaRenderer:
    def __init__(self, env: jinja2.Environment) -> None:
        self._env = env

    def render(self, template_name: str, context: t.Mapping = None) -> str:
        return self._env.get_template(template_name).render(context)


class JinjaProvider(Provider):
    def __init__(
        self,
        template_dirs: list[str] = None,
        tests: t.Mapping[str, t.Callable] = None,
        filters: t.Mapping[str, t.Callable] = None,
        globals: t.Mapping[str, t.Callable] = None,
        policies: t.Mapping[str, t.Callable] = None,
        extensions: t.Sequence[t.Union[str, t.Type[Extension]]] = None,
    ) -> None:
        self.template_dirs = template_dirs or []
        self.filters = filters or {}
        self.globals = globals or {}
        self.tests = tests or {}
        self.policies = policies or {}
        self.extensions = extensions or []

    def register(self, app: Kupala) -> None:
        self.template_dirs = [*self.template_dirs, *app.template_dirs]
        app.services.add_singleton(jinja2.Environment, self.make_environment)
        app.services.add_singleton(TemplateRenderer, self.make_renderer)

    def make_renderer(self, resolver: Resolver) -> TemplateRenderer:
        return JinjaRenderer(env=resolver.resolve(jinja2.Environment))

    def make_environment(self, resolver: Resolver) -> jinja2.Environment:
        env = jinja2.Environment(loader=self.make_loader(self.template_dirs), extensions=self.extensions)
        env.filters.update(self.filters)
        env.globals.update(self.globals)
        env.tests.update(self.tests)
        env.policies.update(self.policies)
        return env

    def make_loader(self, template_dirs: list[str]) -> jinja2.BaseLoader:
        directories = []
        packages = []
        for dir_spec in template_dirs:
            if ':' in dir_spec:
                package_name, package_path = dir_spec.split(':')
                if not package_path:
                    package_path = 'templates'
                packages.append([package_name, package_path])
            else:
                directories.append(dir_spec)

        package_loaders = [jinja2.PackageLoader(package_name, package_path) for package_name, package_path in packages]
        if packages and directories:
            return jinja2.ChoiceLoader(
                [
                    *package_loaders,
                    jinja2.FileSystemLoader(directories),
                ]
            )
        if packages:
            return jinja2.ChoiceLoader(package_loaders)
        return jinja2.FileSystemLoader(directories)



from __future__ import annotations

import copy
import typing as t


class ConfigError(Exception):
    """Base exception for all config related errors."""


class LockedError(ConfigError):
    """Raised when the configuration object is locked."""


class Config:
    """Keeps application configuration three."""

    def __init__(self, initial: t.Mapping = None) -> None:
        self._data = dict(initial or {})
        self._is_locked = False

    def get(self, key: str, default: t.Any = None) -> t.Any:
        """Get a value from the configuration object using dot-notation.
        This method returns a deep copy of the found value."""
        value = self._data
        for segment in key.split("."):
            if segment in value:
                value = value[segment]
            else:
                return default
        return copy.deepcopy(value)

    def get_key(self, key: str, default: t.Any = None) -> t.Any:
        """Get a value by a key."""
        return copy.deepcopy(self._data.get(key, default))

    def set(self, section: str, data: t.Any) -> None:
        """Set a value for a section.
        Raises LockedError if the config object locked."""
        self._raise_if_locked()
        self._data[section] = data

    def update(self, data: t.Any) -> None:
        """Update configuration object.

        This method behaves as dict.update..
        Raises LockedError if the config object locked."""
        self._raise_if_locked()
        self._data.update(data)

    def lock(self) -> None:
        """Lock the config.
        Once locked no config modification allowed.
        Any attempt will raise LockedError."""
        self._is_locked = True

    def unlock(self) -> _Lock:
        """Unlock (or temporary unlock) the config.
        In order to temporary unlock the object
        use it as a context manager:
            with config.unlock() as config:
                config.set('section', 'value')
        """
        self._is_locked = False
        return _Lock(self)

    def _raise_if_locked(self) -> None:
        """Checks if the object is locked and raises LockedError when True."""
        if self._is_locked:
            raise LockedError("The config object is locked. No modifications possible.")

    def __repr__(self) -> str:  # pragma: nocover
        return "<Config: keys=%s>" % self._data.keys()

    __getitem__ = get
    __setitem__ = set
    __call__ = get


class _Lock:
    def __init__(self, config: Config) -> None:
        self._config = config

    def __enter__(self) -> Config:
        return self._config

    def __exit__(self, *args: t.Any) -> None:
        self._config.lock()



from __future__ import annotations

import builtins as bi
import datetime as datetimelib
import decimal
import enum as enumlib
import functools
import json
import logging
import os
import pathlib
import typing as t
import uuid as uuidlib
from dotenv import load_dotenv
from urllib.parse import ParseResult, urlparse

CastFn = t.Callable[[str], t.Any]
ValidatorFn = t.Callable[[t.Any], bool]


class _Undefined:
    ...


undefined = _Undefined()


class EnvError(Exception):
    """Base class for .env related errors."""


class InvalidEnvValueError(ValueError, EnvError):
    """Raised if environment variables does not pass validation."""


def one_of_validator(value: t.Any, choices: list) -> bool:
    return value in choices


def min_max_validator(
    value: t.Union[int, float, decimal.Decimal],
    min: t.Union[int, float, decimal.Decimal] = None,
    min_inclusive: bool = True,
    max: t.Union[int, float, decimal.Decimal] = None,
    max_inclusive: bool = True,
) -> bool:
    checks: list[bool] = []
    if min:
        checks.append(value >= min if min_inclusive else value > min)
    if max:
        checks.append(value <= max if max_inclusive else value < max)
    return all(checks)


class DotEnv:
    def __init__(self, paths: t.Iterable[t.Union[str, os.PathLike[str]]], prefix: bi.str = '') -> None:
        self._prefix = prefix
        for path in paths:
            self.load(path)

    def load(self, file_path: t.Union[str, os.PathLike[str]] = None, stream: t.IO[str] = None) -> DotEnv:
        assert file_path or stream, 'Either "file_path" or "stream" must be passed to "DotEnv.load" method.'
        load_dotenv(file_path, stream)
        return self

    def get(
        self,
        name: str,
        default: t.Any = undefined,
        cast: CastFn = None,
        *,
        validators: list[ValidatorFn] = None,
        check_file: bool = False,
        prefix: bi.str = '',
    ) -> t.Any:
        key = f'{prefix or self._prefix}{name}'
        value = os.environ.get(key, default)
        if value == undefined:
            base_message = f'Environment variable "{name}" is undefined and has no default.'
            if not check_file:
                raise EnvError(base_message)

            try:
                file_path = self.get(f'{key}_FILE')
            except EnvError:
                raise EnvError(
                    f'{base_message} ' f'Also, additional "{key}_FILE" has been searched but is also undefined.'
                )

            try:
                with open(file_path, 'r') as f:
                    value = f.read()
            except FileNotFoundError:
                raise EnvError(
                    f'{base_message} ' f'Also, the file {file_path} specified by "{key}_FILE" variable does not exist.'
                )

        if value != default:
            value = cast(value) if cast else value
        if validators:
            for validator in validators:
                if not validator(value):
                    raise InvalidEnvValueError(f'Environment variable "{key}" has invalid value "{value}".')
        return value

    def int(
        self,
        name: str,
        default: t.Union[int, _Undefined] = undefined,
        *,
        validators: list[ValidatorFn] = None,
        check_file: bool = False,
        prefix: bi.str = '',
        min: int = None,
        min_inclusive: bool = True,
        max: int = None,
        max_inclusive: bool = True,
    ) -> bi.int:
        validators = validators or []
        if min or max:
            validators.append(
                functools.partial(
                    min_max_validator, min=min, min_inclusive=min_inclusive, max=max, max_inclusive=max_inclusive
                )
            )
        return self.get(name, default, int, validators=validators, check_file=check_file, prefix=prefix)

    def float(
        self,
        name: str,
        default: t.Union[float, _Undefined] = undefined,
        *,
        validators: list[ValidatorFn] = None,
        check_file: bool = False,
        prefix: bi.str = '',
        min: float = None,
        min_inclusive: bool = True,
        max: float = None,
        max_inclusive: bool = True,
    ) -> bi.float:
        validators = validators or []
        if min or max:
            validators.append(
                functools.partial(
                    min_max_validator, min=min, min_inclusive=min_inclusive, max=max, max_inclusive=max_inclusive
                )
            )
        return self.get(name, default, float, validators=validators, check_file=check_file, prefix=prefix)

    def decimal(
        self,
        name: str,
        default: t.Union[decimal.Decimal, _Undefined] = undefined,
        *,
        validators: list[ValidatorFn] = None,
        check_file: bool = False,
        prefix: bi.str = '',
        min: decimal.Decimal = None,
        min_inclusive: bool = True,
        max: decimal.Decimal = None,
        max_inclusive: bool = True,
    ) -> decimal.Decimal:
        validators = validators or []
        if min or max:
            validators.append(
                functools.partial(
                    min_max_validator, min=min, min_inclusive=min_inclusive, max=max, max_inclusive=max_inclusive
                )
            )
        return self.get(name, default, decimal.Decimal, validators=validators, check_file=check_file, prefix=prefix)

    def str(
        self,
        name: bi.str,
        default: t.Union[bi.str, _Undefined] = undefined,
        *,
        validators: list[ValidatorFn] = None,
        check_file: bool = False,
        prefix: bi.str = '',
        choices: list[bi.str] = None,
    ) -> bi.str:
        validators = validators or []
        if choices:
            validators.append(functools.partial(one_of_validator, choices=choices))
        return self.get(name, default, str, validators=validators, check_file=check_file, prefix=prefix)

    def bool(
        self,
        name: bi.str,
        default: t.Union[bool, _Undefined] = undefined,
        *,
        validators: list[ValidatorFn] = None,
        check_file: bool = False,
        prefix: bi.str = '',
        allowed_values: list[bi.str] = None,
    ) -> bi.bool:
        allowed_values = allowed_values or ['1', 'true', 't', 'yes', 'on']

        def caster(value: str) -> bi.bool:
            assert allowed_values
            return value.lower() in allowed_values

        return self.get(name, default, caster, validators=validators, check_file=check_file, prefix=prefix)

    def list(
        self,
        name: bi.str,
        default: t.Union[list, _Undefined] = undefined,
        *,
        sub_cast: CastFn = None,
        validators: list[ValidatorFn] = None,
        check_file: bi.bool = False,
        prefix: bi.str = '',
        separator: bi.str = ',',
    ) -> bi.list:
        sub_cast = sub_cast or (lambda x: x)

        def caster(value: str) -> bi.list:
            assert sub_cast
            parts = value.split(separator)
            return list(map(sub_cast, parts))

        return self.get(name, default, caster, validators=validators, check_file=check_file, prefix=prefix)

    def json(
        self,
        name: bi.str,
        default: t.Union[t.Any, _Undefined] = undefined,
        *,
        validators: bi.list[ValidatorFn] = None,
        check_file: bi.bool = False,
        prefix: bi.str = '',
    ) -> t.Any:
        return self.get(name, default, cast=json.loads, validators=validators, check_file=check_file, prefix=prefix)

    def datetime(
        self,
        name: bi.str,
        default: t.Union[datetimelib.datetime, _Undefined] = undefined,
        *,
        validators: bi.list[ValidatorFn] = None,
        check_file: bi.bool = False,
        prefix: bi.str = '',
    ) -> datetimelib.datetime:
        return self.get(
            name,
            default,
            cast=datetimelib.datetime.fromisoformat,
            validators=validators,
            check_file=check_file,
            prefix=prefix,
        )

    def date(
        self,
        name: bi.str,
        default: t.Union[datetimelib.date, _Undefined] = undefined,
        *,
        validators: bi.list[ValidatorFn] = None,
        check_file: bi.bool = False,
        prefix: bi.str = '',
    ) -> datetimelib.date:
        return self.get(
            name,
            default,
            cast=datetimelib.date.fromisoformat,
            validators=validators,
            check_file=check_file,
            prefix=prefix,
        )

    def time(
        self,
        name: bi.str,
        default: t.Union[datetimelib.time, _Undefined] = undefined,
        *,
        validators: bi.list[ValidatorFn] = None,
        check_file: bi.bool = False,
        prefix: bi.str = '',
    ) -> datetimelib.time:
        return self.get(
            name,
            default,
            cast=datetimelib.time.fromisoformat,
            validators=validators,
            check_file=check_file,
            prefix=prefix,
        )

    def timedelta(
        self,
        name: bi.str,
        default: t.Union[datetimelib.timedelta, _Undefined] = undefined,
        *,
        validators: bi.list[ValidatorFn] = None,
        check_file: bi.bool = False,
        prefix: bi.str = '',
    ) -> datetimelib.timedelta:
        def caster(value: str) -> datetimelib.timedelta:
            return datetimelib.timedelta(seconds=int(value))

        return self.get(name, default, cast=caster, validators=validators, check_file=check_file, prefix=prefix)

    def uuid(
        self,
        name: bi.str,
        default: t.Union[uuidlib.UUID, _Undefined] = undefined,
        *,
        validators: bi.list[ValidatorFn] = None,
        check_file: bi.bool = False,
        prefix: bi.str = '',
    ) -> uuidlib.UUID:
        return self.get(name, default, cast=uuidlib.UUID, validators=validators, check_file=check_file, prefix=prefix)

    def url(
        self,
        name: bi.str,
        default: t.Union[bi.str, _Undefined] = undefined,
        *,
        validators: bi.list[ValidatorFn] = None,
        check_file: bi.bool = False,
        prefix: bi.str = '',
    ) -> ParseResult:
        url = self.get(name, default, cast=urlparse, validators=validators, check_file=check_file, prefix=prefix)
        if url == default:
            url = urlparse(url)
        return url

    def log_level(
        self,
        name: bi.str,
        default: t.Union[bi.str, _Undefined] = undefined,
        *,
        validators: bi.list[ValidatorFn] = None,
        check_file: bi.bool = False,
        prefix: bi.str = '',
    ) -> bi.int:
        def caster(value: bi.str) -> bi.int:
            return logging.getLevelName(value)

        return self.get(name, default, cast=caster, validators=validators, check_file=check_file, prefix=prefix)

    def path(
        self,
        name: bi.str,
        default: t.Union[t.Union[bi.str, os.PathLike], _Undefined] = undefined,
        *,
        validators: bi.list[ValidatorFn] = None,
        check_file: bi.bool = False,
        prefix: bi.str = '',
    ) -> pathlib.Path:
        value = self.get(name, default, cast=pathlib.Path, validators=validators, check_file=check_file, prefix=prefix)
        if value == default and isinstance(value, bi.str):
            value = pathlib.Path(value)
        return value

    E = t.TypeVar('E', bound=enumlib.Enum)

    def enum(
        self,
        name: bi.str,
        enum_class: t.Type[E],
        default: t.Union[E, _Undefined] = undefined,
        *,
        validators: bi.list[ValidatorFn] = None,
        check_file: bi.bool = False,
        prefix: bi.str = '',
    ) -> E:
        def caster(value: bi.str) -> t.Any:
            for enum_value in enum_class:
                if enum_value.name == value:
                    return enum_value

        return self.get(name, default, cast=caster, validators=validators, check_file=check_file, prefix=prefix)

    __call__ = str



import typing as t

from kupala import responses
from kupala.container import ServiceNotFoundError
from kupala.contracts import TemplateRenderer
from kupala.requests import Request


def default_app_context(request: Request) -> t.Mapping:
    return {
        'debug': request.app.debug,
        'environment': request.app.env,
        'envvar': request.app.dotenv,
        'url': request.url_for,
        'config': request.app.config,
        'form_errors': request.form_errors,
        'old_input': request.old_input,
    }


class RenderError(Exception):
    """Base class for all renderer classes."""


class TemplateResponse(responses.Response):
    def __init__(
        self,
        request: Request,
        template_name: str,
        context: t.Dict = None,
        status_code: int = 200,
        media_type: str = 'text/html',
        headers: dict = None,
    ) -> None:
        try:
            context = context or {}
            context['request'] = request
            renderer: TemplateRenderer = request.app.resolve(TemplateRenderer)
            context_processors = request.app.context_processors
            for context_processor in context_processors:
                context.update(context_processor(request))
            content = renderer.render(template_name, context)
            super().__init__(content, status_code, headers, media_type)
        except ServiceNotFoundError as ex:
            raise RenderError('A template renderer is not configured for current application.') from ex



from __future__ import annotations

import typing as t
from starlette.types import ASGIApp


class Middleware:
    """Keeps middleware callable along with its constructor arguments."""

    def __init__(self, obj: t.Type, **kwargs: t.Any) -> None:
        self.obj = obj
        self.args = kwargs

    def wrap(self, app: ASGIApp) -> ASGIApp:
        """Wraps ASGI application with this middleware."""
        return self.obj(app, **self.args)

    def __repr__(self) -> str:  # pragma: no cover
        return "<Middleware: %s, kwargs=%r>" % (self.obj, self.args)


class MiddlewareStack:
    """Keeps track about all middleware used."""

    def __init__(self) -> None:
        self._global: list[Middleware] = []

    def top(self, mw: t.Type, **kwargs: t.Any) -> None:
        """Add middleware to the top of stack."""
        self._global.insert(0, Middleware(mw, **kwargs))

    def use(self, mw: t.Type, **kwargs: t.Any) -> None:
        """Add middleware to the end of stack."""
        self._global.append(Middleware(mw, **kwargs))

    def __iter__(self) -> t.Iterator[Middleware]:
        return iter(self._global)

    def __reversed__(self) -> t.Iterator[Middleware]:
        return reversed(self._global)

    def __len__(self) -> int:
        return len(self._global)

    def __str__(self) -> str:  # pragma: nocover
        return "<MiddlewareStack: %s middleware>" % (len(self),)



import abc
import enum
import typing as t
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from kupala.requests import Request

SCOPE_KEY = 'flash_messages'
SESSION_KEY = 'flash_messages'


class MessageCategory(enum.Enum):
    DEBUG = 'debug'
    INFO = 'info'
    SUCCESS = 'success'
    WARNING = 'warning'
    ERROR = 'error'


class FlashMessage(t.TypedDict):
    category: str
    message: str


class FlashBag:
    Category = MessageCategory

    def __init__(self, messages: t.List[FlashMessage] = None):
        self._messages = messages or []

    def add(self, message: str, category: t.Union[MessageCategory, str]) -> None:
        if isinstance(category, MessageCategory):
            category = category.value
        self._messages.append({'category': category, 'message': message})

    def debug(self, message: str) -> None:
        self.add(message, MessageCategory.DEBUG)

    def info(self, message: str) -> None:
        self.add(message, MessageCategory.INFO)

    def success(self, message: str) -> None:
        self.add(message, MessageCategory.SUCCESS)

    def warning(self, message: str) -> None:
        self.add(message, MessageCategory.WARNING)

    def error(self, message: str) -> None:
        self.add(message, MessageCategory.ERROR)

    def all(self) -> t.List[FlashMessage]:
        return self._messages

    def stream(self) -> t.Generator[FlashMessage, None, None]:
        while len(self._messages):
            yield self._messages.pop(0)

    def clear(self) -> None:
        self._messages = []

    def __len__(self) -> int:
        return len(self._messages)


class MessageStorage(abc.ABC):  # pragma: nocover
    @abc.abstractmethod
    def load(self, scope: Scope) -> FlashBag:
        raise NotImplementedError()

    @abc.abstractmethod
    def save(self, scope: Scope, bag: FlashBag) -> None:
        raise NotImplementedError()


class SessionStorage(MessageStorage):
    def load(self, scope: Scope) -> FlashBag:
        if 'session' not in scope:
            raise KeyError('Sessions are disabled. Flash messages depend on SessionMiddleware.')

        return FlashBag(scope['session'].get(SESSION_KEY, []))

    def save(self, scope: Scope, bag: FlashBag) -> None:
        scope['session'][SESSION_KEY] = bag.all()


_storage_map: t.Dict[str, t.Type[MessageStorage]] = {
    'session': SessionStorage,
}


class FlashMessagesMiddleware:
    def __init__(self, app: ASGIApp, storage: t.Union[MessageStorage, t.Literal["session"]] = 'session') -> None:
        self.app = app
        if isinstance(storage, str):
            storage_class = _storage_map[storage]
            storage = storage_class()
        self.storage = storage

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope['type'] != 'http':  # pragma: nocover
            await self.app(scope, receive, send)
            return

        bag = self.storage.load(scope)

        async def send_wrapper(message: Message) -> None:
            if message['type'] == 'http.response.start':
                self.storage.save(scope, bag)
            await send(message)

        scope[SCOPE_KEY] = bag
        await self.app(scope, receive, send_wrapper)


def flash(request: Request) -> FlashBag:
    assert SCOPE_KEY in request.scope, 'Flash messages require FlashMessagesMiddleware.'
    return request.scope[SCOPE_KEY]



import typing as t

from kupala.requests import Request

SERVICE = t.TypeVar('SERVICE')


class Resolver(t.Protocol):  # pragma: nocover
    """A service resolver protocol."""

    def resolve(self, key: t.Type[SERVICE]) -> SERVICE:
        ...


class Invoker(t.Protocol):  # pragma: nocover
    """Invoker is an object that can invoke callables resolving and passing dependencies."""

    def invoke(self, fn_or_class: t.Union[t.Callable, t.Type], extra_kwargs: t.Dict[str, t.Any] = None) -> t.Any:
        ...


class TemplateRenderer(t.Protocol):  # pragma: nocover
    """Render template to string."""

    def render(self, template_name: str, context: t.Dict = None) -> str:
        ...


class ContextProcessor(t.Protocol):  # pragma: nocover
    def __call__(self, request: Request) -> t.Mapping:
        ...



import dataclasses

import datetime
import decimal
import functools
import json
import typing as t
import uuid
from enum import Enum
from starlette.datastructures import FormData

JSONEncoder = json.JSONEncoder


def json_default(o: t.Any) -> t.Any:
    """Usage: json.dumps(data, default=json_default)"""
    if hasattr(o, "to_json"):
        return o.to_json()

    if isinstance(o, (uuid.UUID, datetime.timedelta)):
        return str(o)

    if isinstance(o, datetime.datetime):
        return o.isoformat()

    if isinstance(o, datetime.date):
        return o.isoformat()

    if isinstance(o, datetime.time):
        return o.isoformat()

    if isinstance(o, (set, frozenset)):
        return list(o)

    if isinstance(o, decimal.Decimal):
        return str(o)

    if isinstance(o, Enum):
        return o.value

    if isinstance(o, FormData):
        return dict(o)

    if dataclasses.is_dataclass(o):
        return dataclasses.asdict(o)

    return JSONEncoder().default(o)


def walk(value: t.Any, callback: t.Callable) -> t.Any:
    """Visit each item and apply"""
    if isinstance(value, dict):
        for key, value_ in value.items():
            value[key] = walk(value_, callback)

    if isinstance(value, (list, tuple, set)):
        return list(walk(x, callback) for x in value)

    if dataclasses.is_dataclass(value):
        value = dataclasses.asdict(value)
        return walk(value, callback)

    try:
        return callback(value)
    except TypeError:
        return value


def jsonify(value: t.Any) -> t.Any:
    """Visit each item and convert it into JSON serializable value."""
    return walk(value, functools.partial(json_default))


JSONData = t.Any  # https://github.com/python/typing/issues/182


def dump(
    value: JSONData,
    fp: t.IO[str],
    default: t.Callable[[t.Any], t.Any] = json_default,
    **kwargs: t.Any,
) -> None:
    json.dump(value, fp, default=default, **kwargs)


def dumps(value: JSONData, **kwargs: t.Any) -> str:
    if "default" not in kwargs and "cls" not in kwargs:
        kwargs["default"] = json_default
    return json.dumps(value, **kwargs)


def load(fp: t.IO[str], **kwargs: t.Any) -> str:
    return json.load(fp, **kwargs)


def loads(value: str, **kwargs: t.Any) -> JSONData:
    return json.loads(value, **kwargs)



from __future__ import annotations

import os
import typing as t
import urllib.parse
from starlette import responses
from starlette.background import BackgroundTask
from starlette.datastructures import UploadFile
from starlette.types import Receive, Scope, Send
from urllib.parse import quote

from kupala import json
from kupala.constants import REDIRECT_INPUT_DATA_SESSION_KEY
from kupala.json import JSONEncoder
from kupala.requests import Request

Response = responses.Response
PlainTextResponse = responses.PlainTextResponse
HTMLResponse = responses.HTMLResponse


class JSONResponse(responses.JSONResponse):
    def __init__(
        self,
        content: t.Any,
        status_code: int = 200,
        headers: dict = None,
        indent: int = None,
        default: t.Callable[[t.Any], t.Any] = None,
        encoder_class: t.Type[JSONEncoder] = None,
    ):
        self._indent = indent
        self._encoder_class = encoder_class
        if not encoder_class:
            self._default = default or json.json_default

        super().__init__(content, status_code, headers)

    def render(self, content: t.Any) -> bytes:
        return json.dumps(
            json.jsonify(content),
            ensure_ascii=False,
            allow_nan=False,
            indent=self._indent,
            default=None if self._encoder_class else self._default,
            cls=self._encoder_class,
            separators=(",", ":"),
        ).encode("utf-8")


class FileResponse(responses.FileResponse):
    def __init__(
        self,
        path: t.Union[str, os.PathLike[str]],
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
        file_name: str = None,
        inline: bool = False,
    ):
        headers = headers or {}
        file_name = file_name or os.path.basename(path)
        encoded = urllib.parse.quote_plus(file_name)
        disposition = "inline" if inline else "attachment"
        headers["Content-Disposition"] = f'{disposition}; filename="{encoded}"'
        super().__init__(
            path,
            status_code,
            headers,
            media_type=media_type,
        )


class StreamingResponse(responses.StreamingResponse):
    def __init__(
        self,
        content: t.Any,
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
        file_name: str = None,
        inline: bool = False,
    ):
        disposition = "inline" if inline else "attachment"
        headers = headers or {}
        if file_name:
            headers.update({"content-disposition": f'{disposition}; filename="{file_name}"'})
        super().__init__(
            content=content,
            status_code=status_code,
            headers=headers,
            media_type=media_type,
        )


class RedirectResponse(responses.RedirectResponse):
    def __init__(
        self,
        url: str = None,
        status_code: int = 302,
        headers: dict = None,
        *,
        input_data: t.Mapping = None,
        flash_message: str = None,
        flash_category: str = "info",
        path_name: str = None,
        path_params: dict = None,
    ):
        self.status_code = status_code
        self.body = b''
        self.background: t.Optional[BackgroundTask] = None
        self.init_headers(headers)

        self._url = url
        self._flash_message = flash_message
        self._flash_message_category = flash_category
        self._input_data = input_data
        self._path_name = path_name
        self._path_params = path_params

        assert url or path_name, 'Either "url" or "path_name" argument must be passed.'

    def flash(self, message: str, category: str = 'success') -> RedirectResponse:
        """Set a flash message to the response."""
        self._flash_message = message
        self._flash_message_category = category
        return self

    def with_input(self, input_data: t.Mapping) -> RedirectResponse:
        """Redirect with form input data. Uploaded files will be removed from data."""
        self._input_data = {k: v for k, v in input_data.items() if not isinstance(v, UploadFile)}
        return self

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if self._flash_message and 'flash_messages' in scope:
            scope['flash_messages'].add(self._flash_message, self._flash_message_category)

        if self._input_data and 'session' in scope:
            scope['session'][REDIRECT_INPUT_DATA_SESSION_KEY] = self._input_data

        if self._path_name:
            url = scope['request'].url_for(self._path_name, **(self._path_params or {}))
        else:
            url = self._url

        self.headers["location"] = quote(str(url), safe=":/%#?=@[]!$&'()*+,;")

        return await super().__call__(scope, receive, send)


class EmptyResponse(responses.Response):
    def __init__(self, headers: dict = None) -> None:
        super().__init__(b'', status_code=204, headers=headers)


class GoBackResponse(RedirectResponse):
    def __init__(
        self,
        request: Request,
        flash_message: str = None,
        flash_category: str = 'info',
        input_data: t.Any = None,
        status_code: int = 302,
    ) -> None:
        redirect_to = request.headers.get('referer', '/')
        current_origin = request.url.netloc
        if current_origin not in redirect_to:
            redirect_to = '/'
        super().__init__(
            redirect_to,
            status_code=status_code,
            input_data=input_data,
            flash_message=flash_message,
            flash_category=flash_category,
        )



from __future__ import annotations

import functools
import inspect
import typing
import typing as t
from os import PathLike
from starlette import routing
from starlette.concurrency import run_in_threadpool
from starlette.routing import Mount, WebSocketRoute, compile_path, get_name, iscoroutinefunction_or_partial
from starlette.staticfiles import StaticFiles
from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.exceptions import MethodNotAllowed
from kupala.middleware import Middleware
from kupala.requests import Request
from kupala.resources import get_resource_action
from kupala.responses import RedirectResponse
from kupala.utils import camel_to_snake


def apply_middleware(app: t.Callable, middleware: t.Sequence[Middleware]) -> ASGIApp:
    for mw in reversed(middleware):
        app = mw.wrap(app)
    return app


def request_response(func: typing.Callable) -> ASGIApp:
    """
    Takes a function or coroutine `func(request) -> response`,
    and returns an ASGI application.
    """
    is_coroutine = iscoroutinefunction_or_partial(func)

    # determine request class
    request_class = t.get_type_hints(func).get('request', Request)

    async def app(scope: Scope, receive: Receive, send: Send) -> None:
        request: Request = request_class(scope, receive, send)
        extra_kwargs = {'request': request, **request.path_params}

        if is_coroutine:
            response = await request.app.invoke(func, extra_kwargs=extra_kwargs)
        else:
            response = await run_in_threadpool(request.app.invoke, func, extra_kwargs=extra_kwargs)
        await response(scope, receive, send)

    return app


class Router(routing.Router):
    pass


class Route(routing.Route):
    def __init__(
        self,
        path: str,
        endpoint: typing.Callable,
        *,
        name: str = None,
        include_in_schema: bool = True,
        methods: typing.List[str] = None,
        middleware: t.Sequence[Middleware] = None,
    ) -> None:
        assert path.startswith("/"), "Routed paths must start with '/'"
        self.path = path
        self.endpoint = endpoint
        self.name = get_name(endpoint) if name is None else name
        self.include_in_schema = include_in_schema

        endpoint_handler = endpoint
        while isinstance(endpoint_handler, functools.partial):
            endpoint_handler = endpoint_handler.func
        if inspect.isfunction(endpoint_handler) or inspect.ismethod(endpoint_handler):
            # Endpoint is function or method. Treat it as `func(request) -> response`.
            self.app = request_response(endpoint)
            if methods is None:
                methods = ["GET"]
        else:
            # Endpoint is a class. Treat it as ASGI.
            self.app = endpoint

        if middleware is not None:
            self.app = apply_middleware(t.cast(t.Callable, self.app), middleware)

        if methods is None:
            self.methods = None
        else:
            self.methods = {method.upper() for method in methods}
            if "GET" in self.methods:
                self.methods.add("HEAD")

        self.path_regex, self.path_format, self.param_convertors = compile_path(path)

    def __repr__(self) -> str:
        return f'<Route: path={self.path}, methods={self.methods}, name={self.name}>'


class _RouteAdapter:
    _routes: Routes
    _wrapped_app: t.Optional[ASGIApp]
    _base_app: t.Optional[ASGIApp]
    _middleware: t.Sequence[Middleware]

    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert self._wrapped_app  # silence mypy
        await self._wrapped_app(scope, receive, send)

    def _initialize_apps(self) -> None:
        self._base_app = self._create_base_asgi_app()
        self._wrapped_app = self._apply_middleware(self._base_app)

    def _create_base_asgi_app(self) -> ASGIApp:
        raise NotImplementedError()

    def _apply_middleware(self, app: ASGIApp) -> ASGIApp:
        return apply_middleware(app, self._middleware)

    def __enter__(self) -> Routes:
        self._wrapped_app = None  # force app recreation when routes changed
        return self._routes

    def __exit__(self, *args: t.Any) -> None:
        pass

    def __getattr__(self, item: str) -> t.Any:
        try:
            return getattr(self._base_app, item)
        except AttributeError:
            self._initialize_apps()
            return getattr(self._base_app, item)


class HostRoutes(_RouteAdapter):
    def __init__(
        self,
        host: str,
        routes: t.Union[list[routing.BaseRoute], Routes] = None,
        name: str = None,
        middleware: t.Sequence[Middleware] = None,
    ) -> None:
        self._host = host
        self._name = name
        self._routes = Routes(routes)
        self._middleware = middleware or []
        self._wrapped_app: t.Optional[ASGIApp] = None
        self._base_app: t.Optional[routing.Host] = None

    def _create_base_asgi_app(self) -> ASGIApp:
        return routing.Host(host=self._host, app=Router(routes=self._routes), name=self._name)


class GroupRoutes(_RouteAdapter):
    def __init__(
        self,
        prefix: str,
        routes: t.Union[list[routing.BaseRoute], Routes] = None,
        name: str = None,
        middleware: t.Sequence[Middleware] = None,
    ) -> None:
        self._name = name
        self._prefix = prefix
        self._routes = Routes(routes)
        self._middleware = middleware or []
        self._wrapped_app: t.Optional[ASGIApp] = None
        self._base_app: t.Optional[routing.Mount] = None

    def _apply_middleware(self, app: ASGIApp) -> ASGIApp:
        return app

    def _create_base_asgi_app(self) -> ASGIApp:
        router: ASGIApp = Router(self._routes)
        if self._middleware:
            for mw in reversed(self._middleware):
                router = mw.wrap(router)
        return routing.Mount(path=self._prefix, app=router, name=self._name)

    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert self._base_app  # silence mypy
        await self._base_app.handle(scope, receive, send)


class ResourceRoute(routing.BaseRoute):
    def __init__(
        self,
        path: str,
        resource: object,
        *,
        name: str = None,
        only: list[str] = None,
        exclude: list[str] = None,
        include_in_schema: bool = True,
        middleware: t.Sequence[Middleware] = None,
        redirect_slashes: bool = True,
        id_param: str = 'id:int',
    ) -> None:
        assert path == "" or path.startswith("/"), "Routed paths must start with '/'"
        assert bool(only and exclude) is False, '"exclude" and "only" arguments are mutually exclusive.'

        self.path = path.rstrip("/")
        self.name = name or camel_to_snake(resource.__class__.__name__).replace('_resource', '')
        self.only = only
        self.exclude = exclude
        self.resource = resource
        self.id_param = id_param
        self.middleware = middleware
        self.include_in_schema = include_in_schema

        self.collection_method_map: dict[str, t.Callable] = {
            'head': getattr(self.resource, 'index', None),
            'get': getattr(self.resource, 'index', None),
            'post': getattr(self.resource, 'create', None),
        }
        self.object_method_map: dict[str, t.Callable] = {
            'head': getattr(self.resource, 'show', None),
            'get': getattr(self.resource, 'show', None),
            'put': getattr(self.resource, 'update', None),
            'patch': getattr(self.resource, 'partial_update', getattr(self.resource, 'update', None)),
            'delete': getattr(self.resource, 'destroy', None),
        }

        routes = self._create_routes()
        self.app = Router(routes, redirect_slashes=redirect_slashes)

    def _create_routes(self) -> list[routing.BaseRoute]:
        return [
            *self._get_custom_action_routes(),
            *self._get_edit_routes(),
            *self._get_list_routes(),
            *self._get_object_routes(),
        ]

    @property
    def routes(self) -> list[routing.BaseRoute]:
        return getattr(self.app, "routes", None)

    def matches(self, scope: Scope) -> t.Tuple[routing.Match, Scope]:
        for route in self.routes:
            match, child_scope = route.matches(scope)
            if match != routing.Match.NONE:
                return match, child_scope
        return routing.Match.NONE, {}

    def url_path_for(self, name: str, **path_params: str) -> routing.URLPath:
        return self.app.url_path_for(name, **path_params)

    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:
        await self.app(scope, receive, send)

    def _get_list_routes(self) -> list[routing.BaseRoute]:
        routes: list[routing.BaseRoute] = []
        collection_route_methods: list[str] = []
        if self._is_action_allowed('index'):
            collection_route_methods.extend(['HEAD', 'GET'])
        if self._is_action_allowed('create'):
            collection_route_methods.append('POST')

        if collection_route_methods:
            routes.append(
                Route(
                    self.path,
                    self._list_view,
                    methods=collection_route_methods,
                    name=f'{self.name}.list',
                    include_in_schema=self.include_in_schema,
                    middleware=self.middleware,
                )
            )
        return routes

    def _get_object_routes(self) -> list[routing.BaseRoute]:
        routes: list[routing.BaseRoute] = []
        object_route_methods: list[str] = []
        if self._is_action_allowed('show'):
            object_route_methods.extend(['HEAD', 'GET'])
        if self._is_action_allowed('update'):
            object_route_methods.extend(['PUT', 'PATCH'])
        if self._is_action_allowed('destroy'):
            object_route_methods.append('DELETE')
        if object_route_methods:
            routes.append(
                Route(
                    '%s/{%s}' % (self.path, self.id_param),
                    self._object_view,
                    methods=object_route_methods,
                    name=f'{self.name}.detail',
                    include_in_schema=self.include_in_schema,
                    middleware=self.middleware,
                )
            )
        return routes

    def _get_edit_routes(self) -> list[routing.BaseRoute]:
        routes: list[routing.BaseRoute] = []
        new_action = getattr(self.resource, 'new', None)
        if self._is_action_allowed('new') and new_action:
            routes.append(Route('%s/new' % self.path, new_action, name=f'{self.name}.new', middleware=self.middleware))

        edit_action = getattr(self.resource, 'edit', None)
        if self._is_action_allowed('edit') and edit_action:
            routes.append(
                Route(
                    '%s/{%s}/edit' % (self.path, self.id_param),
                    edit_action,
                    name=f'{self.name}.edit',
                    middleware=self.middleware,
                )
            )
        return routes

    def _get_custom_action_routes(self) -> list[routing.BaseRoute]:
        routes: list[routing.BaseRoute] = []

        for name, fn in inspect.getmembers(self.resource, inspect.ismethod):
            action = get_resource_action(fn)
            if action and self._is_action_allowed(name):
                path = '{base}{id_param}{action}'.format(
                    base=self.path,
                    id_param=('/{%s}' % self.id_param if action.is_for_object else ''),
                    action=action.path,
                )
                routes.append(
                    Route(
                        path=path,
                        endpoint=fn,
                        name=f'{self.name}.{action.path_name}',
                        include_in_schema=action.include_in_schema,
                        methods=action.methods,
                        middleware=action.middleware,
                    )
                )

        return routes

    async def _list_view(self, request: Request) -> ASGIApp:
        action = self.collection_method_map.get(request.method.lower())
        if not action:
            raise MethodNotAllowed()
        return request_response(action)

    async def _object_view(self, request: Request) -> ASGIApp:
        action = self.object_method_map.get(request.method.lower())
        if not action:
            raise MethodNotAllowed()
        return request_response(action)

    def _is_action_allowed(self, action: str) -> bool:
        if self.only:
            return action in self.only
        if self.exclude:
            return action not in self.exclude
        return True


class Routes(t.Sequence[routing.BaseRoute]):
    def __init__(self, routes: t.Iterable[routing.BaseRoute] = None) -> None:
        self._routes: list[routing.BaseRoute] = list(routes or [])

    def get(
        self,
        path: str,
        endpoint: t.Callable,
        *,
        name: str = None,
        include_in_schema: bool = True,
        middleware: t.Sequence[Middleware] = None,
    ) -> None:
        self.add(
            path,
            endpoint,
            name=name,
            methods=['HEAD', 'GET'],
            include_in_schema=include_in_schema,
            middleware=middleware,
        )

    def post(
        self,
        path: str,
        endpoint: t.Callable,
        *,
        name: str = None,
        include_in_schema: bool = True,
        middleware: t.Sequence[Middleware] = None,
    ) -> None:
        self.add(
            path, endpoint, name=name, methods=['POST'], include_in_schema=include_in_schema, middleware=middleware
        )

    def get_or_post(
        self,
        path: str,
        endpoint: t.Callable,
        *,
        name: str = None,
        include_in_schema: bool = True,
        middleware: t.Sequence[Middleware] = None,
    ) -> None:
        self.add(
            path,
            endpoint,
            name=name,
            methods=['HEAD', 'GET', 'POST'],
            include_in_schema=include_in_schema,
            middleware=middleware,
        )

    def patch(
        self,
        path: str,
        endpoint: t.Callable,
        *,
        name: str = None,
        include_in_schema: bool = True,
        middleware: t.Sequence[Middleware] = None,
    ) -> None:
        self.add(
            path, endpoint, name=name, methods=['PATCH'], include_in_schema=include_in_schema, middleware=middleware
        )

    def put(
        self,
        path: str,
        endpoint: t.Callable,
        *,
        name: str = None,
        include_in_schema: bool = True,
        middleware: t.Sequence[Middleware] = None,
    ) -> None:
        self.add(path, endpoint, name=name, methods=['PUT'], include_in_schema=include_in_schema, middleware=middleware)

    def delete(
        self,
        path: str,
        endpoint: t.Callable,
        *,
        name: str = None,
        include_in_schema: bool = True,
        middleware: t.Sequence[Middleware] = None,
    ) -> None:
        self.add(
            path, endpoint, name=name, methods=['DELETE'], include_in_schema=include_in_schema, middleware=middleware
        )

    def head(
        self,
        path: str,
        endpoint: t.Callable,
        *,
        name: str = None,
        include_in_schema: bool = True,
        middleware: t.Sequence[Middleware] = None,
    ) -> None:
        self.add(
            path, endpoint, name=name, methods=['HEAD'], include_in_schema=include_in_schema, middleware=middleware
        )

    def options(
        self,
        path: str,
        endpoint: t.Callable,
        *,
        name: str = None,
        include_in_schema: bool = True,
        middleware: t.Sequence[Middleware] = None,
    ) -> None:
        self.add(
            path, endpoint, name=name, methods=['OPTIONS'], include_in_schema=include_in_schema, middleware=middleware
        )

    def any(
        self,
        path: str,
        endpoint: t.Callable,
        *,
        name: str = None,
        include_in_schema: bool = True,
        middleware: t.Sequence[Middleware] = None,
    ) -> None:
        _all = ["GET", "HEAD", "POST", "PUT", "PATCH", "DELETE", "OPTIONS", "TRACE"]
        self.add(path, endpoint, name=name, methods=_all, include_in_schema=include_in_schema, middleware=middleware)

    def add(
        self,
        path: str,
        endpoint: t.Callable,
        *,
        methods: t.List[str],
        name: str = None,
        include_in_schema: bool = True,
        middleware: t.Sequence[Middleware] = None,
    ) -> None:
        route = Route(
            path,
            endpoint,
            methods=methods,
            name=name,
            include_in_schema=include_in_schema,
            middleware=middleware,
        )
        self._routes.append(route)

    def websocket(self, path: str, endpoint: t.Callable, *, name: str = None) -> None:
        self._routes.append(WebSocketRoute(path, endpoint, name=name))

    def mount(self, path: str, app: ASGIApp, *, name: str = None) -> None:
        self._routes.append(Mount(path, app, name=name))

    def static(
        self,
        path: str,
        directory: t.Union[str, PathLike[str]] = None,
        *,
        packages: list[str] = None,
        html: bool = False,
        check_dir: bool = True,
        name: str = None,
    ) -> None:
        app = StaticFiles(directory=directory, packages=packages, html=html, check_dir=check_dir)
        self._routes.append(Mount(path, app, name=name))

    def host(
        self,
        host: str,
        routes: t.Union[Routes, list[routing.BaseRoute]] = None,
        *,
        name: str = None,
        middleware: t.Sequence[Middleware] = None,
    ) -> HostRoutes:
        app = HostRoutes(host, routes, name, middleware)
        self._routes.append(t.cast(routing.Host, app))
        return app

    def group(
        self,
        prefix: str,
        routes: t.Union[Routes, list[routing.BaseRoute]] = None,
        *,
        name: str = None,
        middleware: t.Sequence[Middleware] = None,
    ) -> GroupRoutes:
        app = GroupRoutes(prefix, routes, name, middleware)
        self._routes.append(t.cast(routing.Mount, app))
        return app

    def redirect(self, path: str, destination: str, status_code: int = 307, headers: dict = None) -> None:
        self.mount(
            path,
            RedirectResponse(destination, status_code=status_code, headers=headers),
        )

    def resource(
        self,
        path: str,
        resource: object,
        *,
        name: str = None,
        only: list[str] = None,
        exclude: list[str] = None,
        include_in_schema: bool = True,
        middleware: t.Sequence[Middleware] = None,
        id_param: str = 'id:int',
    ) -> None:
        self._routes.append(
            ResourceRoute(
                path,
                resource,
                name=name,
                only=only,
                exclude=exclude,
                include_in_schema=include_in_schema,
                middleware=middleware,
                id_param=id_param,
            )
        )

    def api_resource(
        self,
        path: str,
        resource: object,
        *,
        name: str = None,
        only: list[str] = None,
        exclude: list[str] = None,
        include_in_schema: bool = True,
        middleware: t.Sequence[Middleware] = None,
        id_param: str = 'id:int',
    ) -> None:
        excluded = ['new', 'edit']
        only = [action for action in only if action not in excluded] if only else None
        exclude = list({*exclude, *excluded}) if exclude else excluded
        self.resource(
            path,
            resource,
            name=name,
            only=only,
            exclude=exclude,
            include_in_schema=include_in_schema,
            middleware=middleware,
            id_param=id_param,
        )

    def __iter__(self) -> t.Iterator[routing.BaseRoute]:
        return iter(self._routes)

    def __len__(self) -> int:
        return len(self._routes)

    def __getitem__(self, item: t.Any) -> t.Any:  # pragma: nocover
        """Added to implement Sequence protocol."""
        raise NotImplementedError

    def __contains__(self, route: object) -> t.Any:  # pragma: nocover
        raise NotImplementedError



from deesk.storage import Storage

from kupala.application import Kupala
from kupala.container import Resolver
from kupala.providers import Provider
from kupala.storages import Storages


class StoragesProvider(Provider):
    def __init__(self, disks: dict[str, Storage], default: str) -> None:
        self.disks = disks
        self.default = default

    def register(self, app: Kupala) -> None:
        app.services.bind(Storages, Storages(self.disks))
        app.services.factory(Storage, self.get_default_disk)

    def get_default_disk(self, resolver: Resolver) -> Storage:
        return self.disks[self.default]



from __future__ import annotations

import anyio
import contextvars as cv
import logging
import traceback
import typing as t
from contextlib import AsyncExitStack
from starlette.middleware.errors import ServerErrorMiddleware
from starlette.routing import BaseRoute
from starlette.types import ASGIApp, Receive, Scope, Send

from kupala.config import Config
from kupala.container import Container, Resolver
from kupala.contracts import ContextProcessor, Invoker
from kupala.dotenv import DotEnv
from kupala.exceptions import ErrorHandler, ExceptionMiddleware
from kupala.middleware import MiddlewareStack
from kupala.requests import Request
from kupala.routing import Router, Routes

if t.TYPE_CHECKING:
    from kupala.providers import Provider


class Kupala:
    def __init__(
        self,
        debug: bool = None,
        env_prefix: str = '',
        environment: str = None,
        config: t.Mapping = None,
        dotenv: t.List[str] = None,
        template_dirs: list[str] = None,
        providers: t.Iterable['Provider'] = None,
        routes: t.Union[Routes, t.List[BaseRoute]] = None,
        context_processors: t.List[ContextProcessor] = None,
        error_handlers: t.Dict[t.Union[t.Type[Exception], int], ErrorHandler] = None,
    ) -> None:
        self.lifespan: t.List[t.Callable[[Kupala], t.AsyncContextManager]] = []
        self.routes = Routes(list(routes) if routes else [])
        self.config = Config(config)
        self.dotenv = DotEnv(dotenv or ['.env.defaults', '.env'], env_prefix)
        self.env = self.dotenv.str('APP_ENV', 'production') if environment is None else environment
        self.debug = self.dotenv.bool('APP_DEBUG', False) if debug is None else debug
        self.middleware = MiddlewareStack()
        self.providers = providers or []
        self.context_processors: t.List[ContextProcessor] = context_processors or []
        self.error_handlers = error_handlers or {}
        self.services = Container()
        self.template_dirs = template_dirs or []

        self._asgi_app: t.Optional[ASGIApp] = None
        self._router: t.Optional[Router] = None
        self._request_container: cv.ContextVar[Container] = cv.ContextVar('_request_container', default=self.services)
        self._request: cv.ContextVar[t.Optional[Request]] = cv.ContextVar('_current_request')

        self.services.bind(Container, self.services)
        self.services.bind(Resolver, self.services)
        self.services.bind(Invoker, self.services)

    async def lifespan_handler(self, scope: Scope, receive: Receive, send: Send) -> None:
        """ASGI lifespan events handler."""
        self.bootstrap()
        await receive()
        try:
            async with AsyncExitStack() as stack:
                for hook in self.lifespan:
                    await stack.enter_async_context(hook(self))

                await send({"type": "lifespan.startup.complete"})
                try:
                    await receive()
                except anyio.get_cancelled_exc_class():
                    pass
        except BaseException as ex:
            logging.exception(ex)
            text = traceback.format_exc()
            await send({"type": "lifespan.startup.failed", "message": text})
        else:
            await send({"type": "lifespan.shutdown.complete"})

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope["type"] in {"http", "websocket", "lifespan"}
        scope["app"] = self
        set_current_application(self)

        request_container = Container(parent=self.services)
        scope['resolver'] = request_container

        if scope["type"] == "lifespan":
            await self.lifespan_handler(scope, receive, send)
            return

        if scope["type"] == "http":
            request = Request(scope, receive, send)
            request_container.bind(Request, request)
            scope["request"] = request
            self._request.set(request)

        if self._asgi_app is None:
            self._asgi_app = self._create_app()

        self._request_container.set(request_container)
        with request_container.change_context():
            await self._asgi_app(scope, receive, send)

    def cli(self) -> None:
        set_current_application(self)
        self.bootstrap()

    def bootstrap(self) -> None:
        for provider in self.providers:
            provider.register(self)
        for provider in self.providers:
            provider.bootstrap(self.services)

    def _create_app(self) -> ASGIApp:
        app: ASGIApp = Router(routes=self.routes)
        self._router = t.cast(Router, app)
        self.middleware.top(ExceptionMiddleware, handlers=self.error_handlers)
        self.middleware.top(ServerErrorMiddleware, debug=True)
        for mw in reversed(self.middleware):
            app = mw.wrap(app)
        return app

    def invoke(self, fn_or_class: t.Union[t.Callable, t.Type], extra_kwargs: t.Dict[str, t.Any] = None) -> t.Any:
        """Resolve callable dependencies and call it passing dependencies to callable arguments.
        If `fn_or_class` is a type then it will be instantiated.

        Use `extra_kwargs` to provide dependency hints to the injector."""
        return self._request_container.get().invoke(fn_or_class, extra_kwargs=extra_kwargs)

    def resolve(self, key: t.Any) -> t.Any:
        """Resolve a service from the current container."""
        return self._request_container.get().resolve(key)


_current_app: cv.ContextVar[Kupala] = cv.ContextVar('_current_app')


def set_current_application(app: Kupala) -> None:
    _current_app.set(app)


def get_current_application() -> Kupala:
    return _current_app.get()



from deesk.storage import Storage


class Storages:
    def __init__(self, disks: dict[str, Storage]) -> None:
        self._disks = disks

    def get(self, name: str) -> Storage:
        return self._disks[name]



from __future__ import annotations

import mimetypes
import os
import pathlib
import re
import typing
import typing as t
import uuid
from deesk.storage import Storage
from imia import UserToken
from starlette import datastructures as ds, requests
from starlette.requests import empty_receive, empty_send
from starlette.types import Receive, Scope, Send
from starsessions import Session

from .storages import Storages

if t.TYPE_CHECKING:
    from .application import Kupala


class OldFormInput(t.Mapping):
    def __init__(self, data: dict) -> None:
        self._data = data

    def get(self, key: str, default: t.Any = None) -> t.Any:
        return self._data.get(key, default)

    def __getitem__(self, item: str) -> t.Any:
        return self._data[item]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> t.Iterator[t.Tuple[str, t.Any]]:
        return iter(self._data)

    def __contains__(self, item: object) -> bool:
        return item in self._data

    def to_json(self) -> dict:
        return self._data


class FormErrors(t.Mapping):
    def __init__(self, error_message: str, errors: t.Mapping[str, list[str]]) -> None:
        self.message = error_message
        self.field_errors = errors

    def get(self, key: str, default: t.Any = None) -> t.Optional[list[str]]:
        return self.field_errors.get(key, default)

    def to_json(self) -> dict:
        return {
            'message': self.message,
            'field_errors': self.field_errors,
        }

    def keys(self) -> t.AbstractSet[str]:
        return self.field_errors.keys()

    def __bool__(self) -> bool:
        return bool(self.field_errors or self.message)

    def __getitem__(self, item: str) -> list[str]:
        return self.field_errors[item]

    def __contains__(self, item: object) -> bool:
        return item in self.field_errors

    def __iter__(self) -> t.Iterator:
        return iter(self.field_errors)

    def __len__(self) -> int:
        return len(self.field_errors)


class QueryParams(requests.QueryParams):
    def get_bool(self, key: str, default: bool = None) -> t.Optional[bool]:
        value = self.get(key, None)
        return value.lower() in ['t', 'true', 'yes', 'on', '1'] if value is not None else default

    def get_list(self, key: str, subcast: t.Callable = None) -> list[str]:
        items = self.getlist(key)
        if subcast:
            return list(map(subcast, items))
        return items

    def get_int(self, key: str, default: int = None) -> t.Optional[int]:
        value: str = self.get(key, None)
        return int(value) if value is not None and value.isnumeric() else default


class FormData(requests.FormData):
    pass


class UploadFile(ds.UploadFile):
    def __init__(self, request: Request, filename: str, file: typing.IO = None, content_type: str = "") -> None:
        self.request = request
        super().__init__(filename, file, content_type)

    @classmethod
    def from_base_upload_file(cls, upload_file: ds.UploadFile, request: Request) -> UploadFile:
        return UploadFile(
            filename=upload_file.filename,
            file=upload_file.file,
            content_type=upload_file.content_type,
            request=request,
        )

    async def save(self, directory: t.Union[str, os.PathLike], filename: str = None, disk: str = 'default') -> str:
        uploaded_filename = self.filename
        prefix = str(uuid.uuid4().fields[-1])[:5]
        extension = pathlib.Path(uploaded_filename).suffix
        if not extension:
            extension = mimetypes.guess_extension(self.content_type) or 'bin'
        suggested_filename = f'{prefix}_{uploaded_filename}.{extension}'

        file_path = os.path.join(directory, filename or suggested_filename)
        storage: Storage = self.request.app.resolve(Storages).get(disk)
        await storage.put(file_path, await self.read())
        return file_path

    async def read(self, size: int = -1) -> bytes:
        return t.cast(bytes, await super().read(size))

    async def read_string(self, size: int = -1, encoding: str = 'utf8', errors: str = 'strict') -> str:
        return (await self.read(size)).decode(encoding, errors)


class FilesData:
    def __init__(self, files: FormData) -> None:
        self._files = files

    def get(self, key: str, default: t.Any = None) -> t.Optional[UploadFile]:
        return self._files.get(key, default)

    def getlist(self, key: str) -> list[UploadFile]:
        return t.cast(list[UploadFile], self._files.getlist(key))


class Request(requests.Request):
    def __new__(cls, scope: Scope, receive: Receive = empty_receive, send: Send = empty_send) -> Request:
        if 'request' not in scope:
            instance = super().__new__(cls)
            instance.__init__(scope, receive, send)
            scope['request'] = instance
        elif scope['request'].__class__ != cls:
            # view function uses custom request class
            request = scope['request']
            request.__class__ = cls
            scope['request'] = request
        return scope['request']

    @property
    def query_params(self) -> QueryParams:
        return QueryParams(super().query_params)

    @property
    def app(self) -> 'Kupala':
        return self.scope['app']

    @property
    def auth(self) -> UserToken:
        assert "auth" in self.scope, "AuthenticationMiddleware must be installed to access request.auth"
        return self.scope["auth"]

    @property
    def wants_json(self) -> bool:
        """Test if request sends Accept header
        with 'application/json' value."""
        return "application/json" in self.headers.get("accept", "")

    @property
    def is_xhr(self) -> bool:
        """
        Is true when the request is a XMLHttpRequest.
        It works if JavaScript's HTTP client sets an X-Requested-With HTTP header.
        Known frameworks:
        http://en.wikipedia.org/wiki/List_of_Ajax_frameworks#JavaScript
        :return:
        """
        return self.headers.get("x-requested-with", None) == "XMLHttpRequest"

    @property
    def ip(self) -> str:
        """Returns the IP address of user."""
        return self.client.host

    @property
    def secure(self) -> bool:
        """Determine if the request served over SSL."""
        return self.scope["scheme"] == "https"

    @property
    def is_post(self) -> bool:
        """Test if request was made using POST command."""
        return self.method.upper() == "POST"

    @property
    def session(self) -> Session:
        assert "session" in self.scope, "SessionMiddleware must be installed to access request.session"
        return self.scope['session']

    @property
    def old_input(self) -> OldFormInput:
        """Get previous form input. This value does not include uploaded files."""
        if 'session' not in self.scope:
            return OldFormInput({})
        return OldFormInput(self.session.pop('_form_old_input', {}))

    async def remember_form_data(self) -> None:
        """Flush current form data into session so it can be used on the next page to render form errors."""
        if 'session' in self.scope:
            data = await self.form()
            self.session['_form_old_input'] = {k: v for k, v in data.items() if not isinstance(v, ds.UploadFile)}

    @property
    def form_errors(self) -> FormErrors:
        """Get form errors generated on the previous page."""
        if 'session' not in self.scope:
            return FormErrors('', {})
        return FormErrors(self.session.pop('_form_error', ''), self.session.pop('_form_field_errors', {}))

    def set_form_errors(self, error_message: str = '', field_errors: dict = None) -> None:
        """Flush form error message and form field errors into session.
        This data can be later retrieved via `request.form_errors` attribute."""
        if 'session' in self.scope:
            self.session['_form_error'] = error_message
            self.session['_form_field_errors'] = field_errors or {}

    def url_matches(self, *patterns: t.Union[str, t.Pattern]) -> bool:
        for pattern in patterns:
            if pattern == self.url.path:
                return True
            if re.match(pattern, self.url.path):
                return True
        return False

    def full_url_matches(self, *patterns: t.Union[str, t.Pattern]) -> bool:
        for pattern in patterns:
            if pattern == str(self.url):
                return True
            if re.match(pattern, str(self.url)):
                return True
        return False

    async def form(self) -> FormData:
        data = await super().form()
        return FormData(
            [
                (k, UploadFile.from_base_upload_file(v, self) if isinstance(v, ds.UploadFile) else v)
                for k, v in data.multi_items()
            ]
        )

    async def files(self) -> FilesData:
        data = await self.form()
        return FilesData(FormData([(k, v) for k, v in data.multi_items() if isinstance(v, UploadFile)]))



import typing as t
from json import JSONEncoder
from pathlib import Path

from .requests import Request
from .responses import (
    EmptyResponse,
    FileResponse,
    GoBackResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    StreamingResponse,
)
from .templating import TemplateResponse


class ResponseFactory:
    def __init__(self, request: Request, status_code: int = 200, headers: dict = None) -> None:
        self.request = request
        self.status_code = status_code
        self.headers = headers

    def json(
        self,
        data: dict,
        default: t.Callable[[t.Any], t.Any] = None,
        indent: int = None,
        encoder_class: t.Type[JSONEncoder] = None,
    ) -> JSONResponse:
        return JSONResponse(
            data,
            status_code=self.status_code,
            headers=self.headers,
            default=default,
            indent=indent,
            encoder_class=encoder_class,
        )

    def text(self, content: str) -> PlainTextResponse:
        return PlainTextResponse(content, status_code=self.status_code, headers=self.headers)

    def html(self, content: str) -> HTMLResponse:
        return HTMLResponse(content, status_code=self.status_code, headers=self.headers)

    def stream(
        self,
        content: t.Any,
        file_name: str = 'data.bin',
        content_type: str = 'application/octet-stream',
        inline: bool = False,
    ) -> StreamingResponse:
        return StreamingResponse(
            content=content,
            status_code=self.status_code,
            headers=self.headers,
            file_name=file_name,
            inline=inline,
            media_type=content_type,
        )

    def redirect(
        self,
        url: str = None,
        status_code: int = 302,
        *,
        input_data: t.Any = None,
        flash_message: str = None,
        flash_category: str = "info",
        path_name: str = None,
        path_params: dict = None,
    ) -> RedirectResponse:
        return RedirectResponse(
            url=url,
            status_code=status_code,
            headers=self.headers,
            input_data=input_data,
            flash_message=flash_message,
            flash_category=flash_category,
            path_name=path_name,
            path_params=path_params,
        )

    def back(
        self, input_data: t.Any = None, status_code: int = 302, flash_message: str = None, flash_category: str = "info"
    ) -> RedirectResponse:
        return GoBackResponse(
            request=self.request,
            flash_message=flash_message,
            flash_category=flash_category,
            input_data=input_data,
            status_code=status_code,
        )

    def send_file(
        self,
        path: t.Union[str, Path],
        file_name: str,
        content_type: str = 'application/octet-stream',
        inline: bool = False,
    ) -> FileResponse:
        return FileResponse(
            path=path,
            status_code=self.status_code,
            headers=self.headers,
            media_type=content_type,
            file_name=file_name,
            inline=inline,
        )

    def empty(self) -> EmptyResponse:
        return EmptyResponse(headers=self.headers)

    def template(self, template_name: str, context: t.Dict = None, media_type: str = 'text/html') -> TemplateResponse:
        return TemplateResponse(
            request=self.request,
            template_name=template_name,
            context=context,
            status_code=self.status_code,
            headers=self.headers,
            media_type=media_type,
        )



from __future__ import annotations

from dataclasses import dataclass

import abc
import contextvars
import enum
import inspect
import typing as t
import uuid
from contextlib import contextmanager


class ContainerError(Exception):
    """Base class for service container exceptions."""


class ServiceNotFoundError(ContainerError):
    """Raised when service is not registered."""


class InjectionErrorType(enum.Enum):
    POSITIONAL_ONLY_ERROR = 'positional_argument'
    NOT_ANNOTATED_ERROR = 'not_annotated_argument'
    SERVICE_NOT_FOUND = 'service_not_found'


class InjectionError(ContainerError):
    """Raise if injection cannot be proceed."""

    ErrorType = InjectionErrorType

    def __init__(
        self, message: str, error_type: InjectionErrorType, invokation_target: t.Union[t.Callable, t.Type]
    ) -> None:
        class_name = (
            invokation_target.__class__.__name__ if inspect.isclass(invokation_target) else invokation_target.__name__
        )
        addon = f'Tried to invoke: {class_name}{inspect.signature(invokation_target)}.'
        super().__init__(message + f'\n{addon}')
        self.error_type = error_type


S = t.TypeVar('S')
Key = t.Union[str, t.Type[S]]


class Resolver(t.Protocol):  # pragma: nocover
    @t.overload
    def resolve(self, key: t.Type[S]) -> S:
        ...

    @t.overload
    def resolve(self, key: str) -> t.Any:
        ...

    def resolve(self, key: Key) -> t.Any:
        ...

    def __contains__(self, item: Key) -> bool:
        ...


Factory = t.Union[t.Callable[[Resolver], t.Any], type]
Initializer = t.Callable[[Resolver, t.Any], None]


class AbstractFactory(t.Protocol):  # pragma: nocover
    def can_create(self, key: Key) -> bool:
        ...

    def resolve(self, key: Key, resolver: Resolver) -> t.Any:
        ...


class BaseAbstractFactory(abc.ABC):  # pragma: no cover
    def can_create(self, key: Key) -> bool:
        raise NotImplementedError()

    def resolve(self, key: Key, resolver: Resolver) -> t.Any:
        raise NotImplementedError()


class Scope(enum.Enum):
    TRANSIENT = 'transient'
    SINGLETON = 'singleton'
    SCOPED = 'scoped'


class ServiceResolver(abc.ABC):  # pragma: no cover
    def resolve(self, context: dict) -> t.Any:
        raise NotImplementedError()


class InstanceResolver(ServiceResolver):
    def __init__(self, instance: t.Any) -> None:
        self.instance = instance

    def resolve(self, context: dict) -> t.Any:
        return self.instance


class FactoryResolver(ServiceResolver):
    def __init__(
        self,
        resolver: Resolver,
        factory: Factory,
        initializer: Initializer = None,
    ) -> None:
        self.resolver = resolver
        self.factory = factory
        self.initializer = initializer

    def resolve(self, context: dict) -> t.Any:
        instance = self.factory(self.resolver)

        if self.initializer:
            self.initializer(self.resolver, instance)
        return instance


class SingletonResolver(FactoryResolver):
    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self.instance = None

    def resolve(self, context: dict) -> t.Any:
        if not self.instance:
            self.instance = super().resolve(context)
        return self.instance


class ScopedResolver(FactoryResolver):
    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self._resolver_id = uuid.uuid4()

    def resolve(self, context: dict) -> t.Any:
        if self._resolver_id not in context:
            context[self._resolver_id] = super().resolve(context)
        return context[self._resolver_id]


class Container:
    def __init__(self, parent: Resolver = None) -> None:
        self._parent = parent
        self._instances: dict[Key, t.Any] = {}
        self._registry: dict[Key, ServiceResolver] = {}
        self._abs_factories: list[AbstractFactory] = []
        self._abs_factory_cache: dict[Key, AbstractFactory] = {}
        self._context: contextvars.ContextVar[dict] = contextvars.ContextVar('_context', default={})

    def bind(self, key: Key, instance: t.Any) -> None:
        """Add an instance to the container."""
        self._registry[key] = InstanceResolver(instance)

    def factory(
        self,
        key: Key,
        factory: Factory,
        scope: Scope = Scope.TRANSIENT,
        initializer: Initializer = None,
    ) -> None:
        if scope == Scope.SINGLETON:
            self._registry[key] = SingletonResolver(self, factory, initializer)
        elif scope == Scope.SCOPED:
            self._registry[key] = ScopedResolver(self, factory, initializer)
        else:
            self._registry[key] = FactoryResolver(self, factory, initializer)

    def add_singleton(self, key: Key, factory: Factory, initializer: Initializer = None) -> None:
        return self.factory(key, factory, scope=Scope.SINGLETON, initializer=initializer)

    def add_scoped(self, key: Key, factory: Factory, initializer: Initializer = None) -> None:
        return self.factory(key, factory, scope=Scope.SCOPED, initializer=initializer)

    def resolve(self, key: t.Any) -> t.Any:
        """Get a service instance from the container.

        fixme: https://github.com/python/mypy/issues/4717
        """
        service_resolver = self._registry.get(key)
        if service_resolver is None:
            if self._parent is not None:
                return self._parent.resolve(key)

            abs_factory = self._get_abstract_factory_for(key)
            if abs_factory is None:
                raise ServiceNotFoundError('Service "%s" is not registered.' % key)
            return abs_factory.resolve(key, self)
        return service_resolver.resolve(self._context.get())

    def invoke(self, fn_or_class: t.Union[t.Callable, t.Type], extra_kwargs: t.Dict[str, t.Any] = None) -> t.Any:
        """Invoke a callable resolving and injecting dependencies."""
        injections = {}
        extra_kwargs = extra_kwargs or {}

        for argument in get_callable_types(fn_or_class):
            if argument.name in extra_kwargs:
                injections[argument.name] = extra_kwargs[argument.name]
                continue

            if argument.is_positional_only:
                raise InjectionError(
                    'Position-only arguments are not supported. ', InjectionErrorType.POSITIONAL_ONLY_ERROR, fn_or_class
                )
            if not argument.is_annotated:
                raise InjectionError(
                    f'Argument "{argument.name}" is not annotated thus injector cannot resolve object to inject. ',
                    InjectionErrorType.NOT_ANNOTATED_ERROR,
                    fn_or_class,
                )

            try:
                injections[argument.name] = self.resolve(argument.annotation)
            except ServiceNotFoundError as ex:
                raise InjectionError(
                    f'Argument "{argument.name}" refers to unregistered service "{argument.annotation}". ',
                    InjectionErrorType.SERVICE_NOT_FOUND,
                    fn_or_class,
                ) from ex

        if inspect.iscoroutinefunction(fn_or_class):

            async def wrapper() -> t.Any:
                return await fn_or_class(**injections)

            return wrapper()
        else:
            return fn_or_class(**injections)

    def add_abstract_factory(self, factory: AbstractFactory) -> None:
        self._abs_factories.append(factory)

    @contextmanager
    def change_context(self, context_data: t.Dict = None) -> t.Generator[Container, None, None]:
        context = self._context.get()
        self._context.set(context_data or {})
        try:
            yield self
        finally:
            self._context.set(context)

    def _get_abstract_factory_for(self, key: Key) -> t.Optional[AbstractFactory]:
        if key not in self._abs_factory_cache:
            for factory in self._abs_factories:
                if factory.can_create(key):
                    self._abs_factory_cache[key] = factory
                    break
        return self._abs_factory_cache.get(key)

    __setitem__ = bind
    __getitem__ = resolve

    def __contains__(self, item: Key) -> bool:
        has_in_self = item in self._registry
        has_in_parent = item in self._parent if self._parent else False
        return any([has_in_self, has_in_parent])


undefined = object()


@dataclass
class _Argument:
    name: str
    kind: int
    annotation: t.Type
    default: t.Any = undefined

    @property
    def has_default(self) -> bool:
        return self.default is not undefined

    @property
    def is_annotated(self) -> bool:
        return self.annotation is not undefined

    @property
    def is_positional_only(self) -> bool:
        return self.kind == inspect.Parameter.POSITIONAL_ONLY

    @property
    def is_keyword(self) -> bool:
        return self.kind in [inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD]


@dataclass
class FunctionTypes:
    arguments: dict[str, _Argument]
    return_type: t.Type

    @property
    def is_return_type_annotated(self) -> bool:
        return self.return_type is not undefined

    def __iter__(self) -> t.Iterator[_Argument]:
        return iter(self.arguments.values())

    def __len__(self) -> int:
        return len(self.arguments)

    def __contains__(self, item: str) -> bool:
        return item in self.arguments

    def __getitem__(self, item: str) -> _Argument:
        return self.arguments[item]


def get_callable_types(fn: t.Union[t.Callable, t.Type]) -> FunctionTypes:
    for_checking = []
    is_class = inspect.isclass(fn)
    if is_class:
        for base in getattr(fn, "__mro__", []):
            if base != object:
                for_checking.append(getattr(base, "__init__"))
    else:
        for_checking.append(fn)

    arguments = {}

    def _get_annotation(param: inspect.Parameter, fn_to_check: t.Callable) -> t.Any:
        if param.annotation == inspect.Parameter.empty:
            return undefined

        type_hints = t.get_type_hints(fn_to_check)
        return type_hints[param.name]

    return_annotation: t.Any = undefined
    for fn_to_check in reversed(for_checking):
        signature = inspect.signature(fn_to_check)
        return_annotation = t.get_type_hints(fn_to_check).get('return', undefined)

        for name, param in signature.parameters.items():
            if param.kind in [inspect.Parameter.VAR_KEYWORD, inspect.Parameter.VAR_POSITIONAL]:
                # ignore *args, **kwargs
                continue
            arguments[name] = _Argument(
                name=name,
                kind=param.kind,
                annotation=_get_annotation(param, fn_to_check),
                default=undefined if param.default == signature.empty else param.default,
            )

    if is_class and 'self' in arguments:
        del arguments['self']

    return FunctionTypes(arguments, return_annotation)
